import re
from typing import Any, Dict, List

contractions = [
    {"contraction": "'bout", "full_forms": ["about"], "type": "word"},
    {"contraction": "'cause", "full_forms": ["because"], "type": "word"},
    {"contraction": "'cept", "full_forms": ["except"], "type": "word"},
    {"contraction": "'em", "full_forms": ["them"], "type": "word"},
    {"contraction": "'gainst", "full_forms": ["against"], "type": "word"},
    {"contraction": "'n'", "full_forms": ["and"], "type": "word"},
    {"contraction": "'neath", "full_forms": ["beneath"], "type": "word"},
    {"contraction": "'round", "full_forms": ["around"], "type": "word"},
    {"contraction": "'thout", "full_forms": ["without"], "type": "word"},
    {"contraction": "'til", "full_forms": ["until"], "type": "word"},
    {"contraction": "'tis", "full_forms": ["it is"], "type": "word"},
    {"contraction": "'twas", "full_forms": ["it was"], "type": "word"},
    {"contraction": "'tween", "full_forms": ["between"], "type": "word"},
    {"contraction": "'twere", "full_forms": ["it were"], "type": "word"},
    {"contraction": "I'd", "full_forms": ["I had", "I would"], "type": "word"},
    {
        "contraction": "I'd'nt",
        "full_forms": ["I would not", "I did not"],
        "type": "word",
    },
    {
        "contraction": "I'd'nt've",
        "full_forms": ["I would not have", "I did not have"],
        "type": "word",
    },
    {
        "contraction": "I'd've",
        "full_forms": ["I would have", "I did have"],
        "type": "word",
    },
    {"contraction": "I'll", "full_forms": ["I shall", "I will"], "type": "word"},
    {"contraction": "I'm", "full_forms": ["I am"], "type": "word"},
    {"contraction": "I'm'na", "full_forms": ["I am going to"], "type": "word"},
    {"contraction": "I'm'o", "full_forms": ["I am going to"], "type": "word"},
    {"contraction": "I'm'onna", "full_forms": ["I am going to"], "type": "word"},
    {"contraction": "I've", "full_forms": ["I have"], "type": "word"},
    {"contraction": "Idunno", "full_forms": ["I do not know"], "type": "word"},
    {"contraction": "If'n", "full_forms": ["If and when"], "type": "word"},
    {
        "contraction": "Imma",
        "full_forms": ["I am about to", "I am going to"],
        "type": "word",
    },
    {"contraction": "Ion", "full_forms": ["I do not", "I do not"], "type": "word"},
    {"contraction": "a'ight", "full_forms": ["alright"], "type": "word"},
    {
        "contraction": "ain't",
        "full_forms": ["am not", "is not", "are not", "has not", "have not", "did not"],
        "type": "word",
    },
    {"contraction": "amn't", "full_forms": ["am not"], "type": "word"},
    {"contraction": "aren't", "full_forms": ["are not"], "type": "word"},
    {
        "contraction": "arencha",
        "full_forms": ["are not you", "are not you"],
        "type": "word",
    },
    {"contraction": "c'mon", "full_forms": ["come on"], "type": "word"},
    {"contraction": "can't", "full_forms": ["cannot"], "type": "word"},
    {"contraction": "cap'n", "full_forms": ["captain"], "type": "word"},
    {"contraction": "could've", "full_forms": ["could have"], "type": "word"},
    {"contraction": "couldn't", "full_forms": ["could not"], "type": "word"},
    {"contraction": "couldn't've", "full_forms": ["could not have"], "type": "word"},
    {"contraction": "cuppa", "full_forms": ["cup of"], "type": "word"},
    {"contraction": "cuz", "full_forms": ["because"], "type": "word"},
    {"contraction": "d'ya", "full_forms": ["do you", "did you"], "type": "word"},
    {"contraction": "d'ye", "full_forms": ["do you", "did you"], "type": "word"},
    {"contraction": "daren't", "full_forms": ["dare not", "dared not"], "type": "word"},
    {"contraction": "daresn't", "full_forms": ["dare not"], "type": "word"},
    {"contraction": "dasn't", "full_forms": ["dare not"], "type": "word"},
    {"contraction": "didn't", "full_forms": ["did not"], "type": "word"},
    {"contraction": "doesn't", "full_forms": ["does not"], "type": "word"},
    {"contraction": "don't", "full_forms": ["do not", "does not"], "type": "word"},
    {
        "contraction": "dunno",
        "full_forms": ["do not know", "do not know"],
        "type": "word",
    },
    {"contraction": "e'en", "full_forms": ["even"], "type": "word"},
    {"contraction": "e'er", "full_forms": ["ever"], "type": "word"},
    {
        "contraction": "everybody's",
        "full_forms": ["everybody has", "everybody is"],
        "type": "word",
    },
    {
        "contraction": "everyone's",
        "full_forms": ["everyone has", "everyone is"],
        "type": "word",
    },
    {
        "contraction": "everything's",
        "full_forms": ["everything has", "everything is"],
        "type": "word",
    },
    {"contraction": "finna", "full_forms": ["fixing to"], "type": "word"},
    {"contraction": "fo'c'sle", "full_forms": ["forecastle"], "type": "word"},
    {"contraction": "g'day", "full_forms": ["good day"], "type": "word"},
    {"contraction": "gi'z", "full_forms": ["give us"], "type": "word"},
    {"contraction": "gimme", "full_forms": ["give me"], "type": "word"},
    {"contraction": "giv'n", "full_forms": ["given"], "type": "word"},
    {"contraction": "gon't", "full_forms": ["go not"], "type": "word"},
    {"contraction": "gonna", "full_forms": ["going to"], "type": "word"},
    {"contraction": "gotta", "full_forms": ["got to"], "type": "word"},
    {"contraction": "had've", "full_forms": ["had have"], "type": "word"},
    {"contraction": "hadn't", "full_forms": ["had not"], "type": "word"},
    {"contraction": "hasn't", "full_forms": ["has not"], "type": "word"},
    {"contraction": "haven't", "full_forms": ["have not"], "type": "word"},
    {"contraction": "he'd", "full_forms": ["he had", "he would"], "type": "word"},
    {
        "contraction": "he'd'nt've",
        "full_forms": ["he did not have", "he would not have"],
        "type": "word",
    },
    {"contraction": "he'll", "full_forms": ["he shall", "he will"], "type": "word"},
    {"contraction": "he's", "full_forms": ["he has", "he is"], "type": "word"},
    {"contraction": "helluva", "full_forms": ["hell of a"], "type": "word"},
    {"contraction": "here's", "full_forms": ["here is"], "type": "word"},
    {"contraction": "how'd", "full_forms": ["how did", "how would"], "type": "word"},
    {"contraction": "how'll", "full_forms": ["how will", "how shall"], "type": "word"},
    {"contraction": "how're", "full_forms": ["how are"], "type": "word"},
    {
        "contraction": "how's",
        "full_forms": ["how has", "how is", "how does"],
        "type": "word",
    },
    {
        "contraction": "howdy",
        "full_forms": ["how do you do", "how do you fare"],
        "type": "word",
    },
    {"contraction": "innit", "full_forms": ["is not it", "am not it"], "type": "word"},
    {"contraction": "isn't", "full_forms": ["is not"], "type": "word"},
    {"contraction": "it'd", "full_forms": ["it would"], "type": "word"},
    {"contraction": "it'll", "full_forms": ["it shall", "it will"], "type": "word"},
    {"contraction": "it's", "full_forms": ["it has", "it is"], "type": "word"},
    {"contraction": "kinda", "full_forms": ["kind of"], "type": "word"},
    {"contraction": "lemme", "full_forms": ["let me"], "type": "word"},
    {"contraction": "let's", "full_forms": ["let us"], "type": "word"},
    {"contraction": "loven't", "full_forms": ["love not"], "type": "word"},
    {"contraction": "ma'am", "full_forms": ["madam"], "type": "word"},
    {"contraction": "may've", "full_forms": ["may have"], "type": "word"},
    {"contraction": "mayn't", "full_forms": ["may not"], "type": "word"},
    {"contraction": "methinks", "full_forms": ["I think"], "type": "word"},
    {"contraction": "might've", "full_forms": ["might have"], "type": "word"},
    {"contraction": "mightn't", "full_forms": ["might not"], "type": "word"},
    {"contraction": "mine's", "full_forms": ["mine is"], "type": "word"},
    {"contraction": "must've", "full_forms": ["must have"], "type": "word"},
    {"contraction": "mustn't", "full_forms": ["must not"], "type": "word"},
    {"contraction": "mustn't've", "full_forms": ["must not have"], "type": "word"},
    {"contraction": "nal", "full_forms": ["and all"], "type": "word"},
    {"contraction": "ne'er", "full_forms": ["never"], "type": "word"},
    {"contraction": "needn't", "full_forms": ["need not"], "type": "word"},
    {"contraction": "o'clock", "full_forms": ["of the clock"], "type": "word"},
    {"contraction": "o'er", "full_forms": ["over"], "type": "word"},
    {"contraction": "ol'", "full_forms": ["old"], "type": "word"},
    {"contraction": "ought've", "full_forms": ["ought have"], "type": "word"},
    {"contraction": "oughtn't", "full_forms": ["ought not"], "type": "word"},
    {"contraction": "oughtn't've", "full_forms": ["ought not have"], "type": "word"},
    {"contraction": "shalln't", "full_forms": ["shall not"], "type": "word"},
    {"contraction": "shan'", "full_forms": ["shall not"], "type": "word"},
    {"contraction": "shan't", "full_forms": ["shall not"], "type": "word"},
    {"contraction": "she'd", "full_forms": ["she had", "she would"], "type": "word"},
    {
        "contraction": "she'd'nt've",
        "full_forms": ["she did not have", "she would not have"],
        "type": "word",
    },
    {"contraction": "she'll", "full_forms": ["she shall", "she will"], "type": "word"},
    {"contraction": "she's", "full_forms": ["she has", "she is"], "type": "word"},
    {"contraction": "should've", "full_forms": ["should have"], "type": "word"},
    {"contraction": "shouldn't", "full_forms": ["should not"], "type": "word"},
    {"contraction": "shouldn't've", "full_forms": ["should not have"], "type": "word"},
    {"contraction": "so're", "full_forms": ["so are"], "type": "word"},
    {"contraction": "so's", "full_forms": ["so is", "so has"], "type": "word"},
    {"contraction": "so've", "full_forms": ["so have"], "type": "word"},
    {
        "contraction": "somebody's",
        "full_forms": ["somebody has", "somebody is"],
        "type": "word",
    },
    {
        "contraction": "someone's",
        "full_forms": ["someone has", "someone is"],
        "type": "word",
    },
    {
        "contraction": "something's",
        "full_forms": ["something has", "something is"],
        "type": "word",
    },
    {"contraction": "that'd", "full_forms": ["that would", "that had"], "type": "word"},
    {
        "contraction": "that'll",
        "full_forms": ["that shall", "that will"],
        "type": "word",
    },
    {"contraction": "that're", "full_forms": ["that are"], "type": "word"},
    {"contraction": "that's", "full_forms": ["that has", "that is"], "type": "word"},
    {
        "contraction": "there'd",
        "full_forms": ["there had", "there would"],
        "type": "word",
    },
    {
        "contraction": "there'll",
        "full_forms": ["there shall", "there will"],
        "type": "word",
    },
    {"contraction": "there're", "full_forms": ["there are"], "type": "word"},
    {"contraction": "there's", "full_forms": ["there has", "there is"], "type": "word"},
    {"contraction": "these're", "full_forms": ["these are"], "type": "word"},
    {"contraction": "these've", "full_forms": ["these have"], "type": "word"},
    {"contraction": "they'd", "full_forms": ["they had", "they would"], "type": "word"},
    {
        "contraction": "they'd've",
        "full_forms": ["they would have", "they could have", "they should have"],
        "type": "word",
    },
    {
        "contraction": "they'll",
        "full_forms": ["they shall", "they will"],
        "type": "word",
    },
    {"contraction": "they're", "full_forms": ["they are", "they were"], "type": "word"},
    {"contraction": "they've", "full_forms": ["they have"], "type": "word"},
    {"contraction": "this's", "full_forms": ["this has", "this is"], "type": "word"},
    {"contraction": "those're", "full_forms": ["those are"], "type": "word"},
    {"contraction": "those've", "full_forms": ["those have"], "type": "word"},
    {"contraction": "to've", "full_forms": ["to have"], "type": "word"},
    {"contraction": "tryna", "full_forms": ["trying to"], "type": "word"},
    {"contraction": "w'all", "full_forms": ["we all"], "type": "word"},
    {"contraction": "w'at", "full_forms": ["we at"], "type": "word"},
    {"contraction": "w/", "full_forms": ["with"], "type": "word"},
    {"contraction": "w/o", "full_forms": ["without"], "type": "word"},
    {"contraction": "wanna", "full_forms": ["want to"], "type": "word"},
    {"contraction": "wasn't", "full_forms": ["was not"], "type": "word"},
    {
        "contraction": "we'd",
        "full_forms": ["we had", "we would", "we did"],
        "type": "word",
    },
    {"contraction": "we'd've", "full_forms": ["we would have"], "type": "word"},
    {"contraction": "we'll", "full_forms": ["we shall", "we will"], "type": "word"},
    {"contraction": "we're", "full_forms": ["we are"], "type": "word"},
    {"contraction": "we've", "full_forms": ["we have"], "type": "word"},
    {"contraction": "weren't", "full_forms": ["were not"], "type": "word"},
    {"contraction": "what'd", "full_forms": ["what did"], "type": "word"},
    {
        "contraction": "what'll",
        "full_forms": ["what shall", "what will"],
        "type": "word",
    },
    {"contraction": "what're", "full_forms": ["what are", "what were"], "type": "word"},
    {
        "contraction": "what's",
        "full_forms": ["what has", "what is", "what does"],
        "type": "word",
    },
    {"contraction": "what've", "full_forms": ["what have"], "type": "word"},
    {"contraction": "whatcha", "full_forms": ["what are you"], "type": "word"},
    {"contraction": "when'd", "full_forms": ["when did"], "type": "word"},
    {"contraction": "when's", "full_forms": ["when has", "when is"], "type": "word"},
    {"contraction": "where'd", "full_forms": ["where did"], "type": "word"},
    {
        "contraction": "where'll",
        "full_forms": ["where shall", "where will"],
        "type": "word",
    },
    {"contraction": "where're", "full_forms": ["where are"], "type": "word"},
    {
        "contraction": "where's",
        "full_forms": ["where has", "where is", "where does"],
        "type": "word",
    },
    {"contraction": "where've", "full_forms": ["where have"], "type": "word"},
    {
        "contraction": "which'd",
        "full_forms": ["which had", "which would"],
        "type": "word",
    },
    {
        "contraction": "which'll",
        "full_forms": ["which shall", "which will"],
        "type": "word",
    },
    {"contraction": "which're", "full_forms": ["which are"], "type": "word"},
    {"contraction": "which's", "full_forms": ["which has", "which is"], "type": "word"},
    {"contraction": "which've", "full_forms": ["which have"], "type": "word"},
    {
        "contraction": "who'd",
        "full_forms": ["who would", "who had", "who did"],
        "type": "word",
    },
    {"contraction": "who'd've", "full_forms": ["who would have"], "type": "word"},
    {"contraction": "who'll", "full_forms": ["who shall", "who will"], "type": "word"},
    {"contraction": "who're", "full_forms": ["who are"], "type": "word"},
    {
        "contraction": "who's",
        "full_forms": ["who has", "who is", "who does"],
        "type": "word",
    },
    {"contraction": "who've", "full_forms": ["who have"], "type": "word"},
    {"contraction": "why'd", "full_forms": ["why did"], "type": "word"},
    {"contraction": "why're", "full_forms": ["why are"], "type": "word"},
    {
        "contraction": "why's",
        "full_forms": ["why has", "why is", "why does"],
        "type": "word",
    },
    {"contraction": "willn't", "full_forms": ["will not"], "type": "word"},
    {"contraction": "won't", "full_forms": ["will not"], "type": "word"},
    {"contraction": "wonnot", "full_forms": ["will not"], "type": "word"},
    {"contraction": "would've", "full_forms": ["would have"], "type": "word"},
    {"contraction": "wouldn't", "full_forms": ["would not"], "type": "word"},
    {"contraction": "wouldn't've", "full_forms": ["would not have"], "type": "word"},
    {
        "contraction": "y'ain't",
        "full_forms": ["you are not", "you have not", "you did not"],
        "type": "word",
    },
    {"contraction": "y'all", "full_forms": ["you all"], "type": "word"},
    {"contraction": "y'all'd've", "full_forms": ["you all would have"], "type": "word"},
    {
        "contraction": "y'all'dn't've",
        "full_forms": ["you all would not have"],
        "type": "word",
    },
    {"contraction": "y'all're", "full_forms": ["you all are"], "type": "word"},
    {"contraction": "y'all'ren't", "full_forms": ["you all are not"], "type": "word"},
    {"contraction": "y'at", "full_forms": ["you at"], "type": "word"},
    {"contraction": "y'ever", "full_forms": ["have you ever"], "type": "word"},
    {"contraction": "y'know", "full_forms": ["you know"], "type": "word"},
    {"contraction": "yes'm", "full_forms": ["yes madam", "yes madam"], "type": "word"},
    {"contraction": "yesn't", "full_forms": ["yes not", "yes no"], "type": "word"},
    {"contraction": "yessir", "full_forms": ["yes sir"], "type": "word"},
    {"contraction": "you'd", "full_forms": ["you had", "you would"], "type": "word"},
    {"contraction": "you'll", "full_forms": ["you shall", "you will"], "type": "word"},
    {"contraction": "you're", "full_forms": ["you are"], "type": "word"},
    {"contraction": "you've", "full_forms": ["you have"], "type": "word"},
    {
        "contraction": "'s",
        "full_forms": ["is", "has", "does", "us", " is"],
        "type": "suffix",
    },
    {"contraction": "'ll", "full_forms": ["will"], "type": "suffix"},
    {"contraction": "'d", "full_forms": ["had", "would"], "type": "suffix"},
    {"contraction": "'re", "full_forms": ["are", "were"], "type": "suffix"},
    {"contraction": "'m", "full_forms": ["am"], "type": "suffix"},
    {"contraction": "n't", "full_forms": ["not"], "type": "suffix"},
]


def build_contractions_regex(contractions: List[Dict[str, Any]]):
    """Build a regex pattern for contractions."""
    cre = []
    for i, v in enumerate(contractions):
        c = v["contraction"]
        if v["type"] == "word":
            cre.append(rf"(?:\b(?P<C{i}>" + re.escape(c) + r")\b)")
    for i, v in enumerate(contractions):
        c = v["contraction"]
        if v["type"] == "suffix":
            cre.append(rf"(?:(?P<C{i}>" + re.escape(c) + r")\b)")
    return re.compile(r"(?:" + "|".join(cre) + r")", re.VERBOSE | re.I | re.UNICODE)


regex = build_contractions_regex(contractions)


# uncontract
def expand(text: str):
    def replace(match: re.Match):
        info = contractions[int(match.lastgroup[1:])]
        if info["type"] == "word":
            return info["full_forms"][0]
        elif info["type"] == "suffix":
            return " " + info["full_forms"][0]
        return match.group(0)

    return regex.sub(replace, text)
